"""
ECCO v4 Python: read_bin_llc

This module includes utility routines for loading binary files in the 
llc 13-tile native flat binary layout.  This layout is the default for 
MITgcm input and output for global setups using lat-lon-cap (llc) layout. 
The llc layout is used for ECCO v4. 

.. _ecco_v4_py Documentation :
   https://github.com/ECCO-GROUP/ECCOv4-py

"""

from __future__ import division,print_function
import xmitgcm

from .llc_array_conversion  import llc_compact_to_tiles, \
    llc_compact_to_faces, llc_faces_to_tiles, llc_faces_to_compact, \
    llc_tiles_to_faces, llc_tiles_to_compact

from .ecco_utils import make_time_bounds_and_center_times_from_ecco_dataset


def load_ecco_vars_from_mds(mds_var_dir, 
                            mds_files, 
                            mds_grid_dir,
                            vars_to_load = 'all', 
                            tiles_to_load = range(13),
                            model_time_steps_to_load = 'all',
                            output_freq_code = '', 
                            meta_variable_specific=dict(),
                            meta_common=dict(),
                            mds_datatype = '>f4'):
                                 
    """

    Uses xmitgcm's *open_mdsdataset* routine to load ecco variable(s) from 
    MITgcm's MDS binary output into xarray Dataset/DataArray objects.  

    The main benefit of using this routine over open_mdsdataset is that this
    routine allows for 
        
        - proper centering of the *time* variable for time-averaged fields

        - creation of the *time-bnds* fields in time-averaged fields

        - specification of extra variable-specific and globl metadata


    xmitgcm.open_mdsdataset uses the model step number from the file name 
    (e.g., 732 from the file VAR.000000000732.data) to construct the 
    'time' field.  For time-averaged fields, this model step 
    corresponds to END of the averaging period, not the time averaging mid 
    point. This routine fixes the 'time' field of time-averaged fields
    to be the mid point of the time averaging period when the appropriate
    *output_freq_code* is passed.

    
    Parameters
    ----------
    mds_var_dir : str
        directory where the .data/.meta files are stored
    
    mds_files   : str or list
        a string or list of file names to load.  
        Note :  the name is everything BEFORE the time step
        the mds_file name for 'var.000000000732.data' is 'var'
    
    mds_grid_dir : str
        the directory where the model binary (.data) grid fields
        are stored
    
    vars_to_load : str or list, optional, default 'all'
        a string or list of the variable names to read from the mds_files

        - if 'all' then all variables in the files are loaded

    tiles_to_load : int or list of ints, optional, default range(13)
        an int or list of ints indicating which tiles to load

    model_time_steps_to_load : int or list of ints, optional, default 'all'
        an int or list of ints indicating which model time steps to load

        Note : the model time step indicates the time step when the the file was written.
        when the field is a time average, this time step shows the END of the averaging period.

    output_freq_code : str, optional, default empty string
        a code used to create the proper time indices on the fields after loading
        ('AVG' or 'SNAPSHOT') + '_' + ('DAY','WEEK','MON', or 'YEAR')
        
        valid options :

        - AVG_DAY, AVG_WEEK, AVG_MON, AVG_YEAR
        - SNAPSHOT_DAY, SNAPSHOT_WEEK, SNAPSHOT_MON, SNAPSHOT_YEAR
        
    meta_variable_specific : dict, optional, default empty dictionary
        a dictionary with variable-specific metadata.  used when creating
        the offical ECCO products

    meta_common : dict, optional, default empty dictionary
        a dictionary with globally-valid metadata for the ECCO fields.
        useful when creating the offical ECCO netcdf fields

    mds_datatype : string, optional, default '>f4'
        code indicating what type of field to load if the xmitgcm cannot 
        determine the format from the .meta file.  '>f4' means big endian
        32 bit float.


    Returns
    =======

    ecco_dataset : xarray Dataset


    """


    #ECCO v4 r3 starts 1992/1/1 12:00:00
    ecco_v4_start_year = 1992
    ecco_v4_start_mon  = 1
    ecco_v4_start_day  = 1
    ecco_v4_start_hour = 12
    ecco_v4_start_min  = 0
    ecco_v4_start_sec  = 0
    
    # ECCO v4 r3 has 1 hour (3600 s) time steps    
    delta_t = 3600
    
    # define reference date for xmitgcm
    ref_date = str(ecco_v4_start_year) + '-' + str(ecco_v4_start_mon)  + '-' + \
        str(ecco_v4_start_day)  + ' ' +  str(ecco_v4_start_hour) + ':' +  \
        str(ecco_v4_start_min)  + ':' + str(ecco_v4_start_sec)


    if model_time_steps_to_load == 'all':

        ecco_dataset = open_mdsdataset(data_dir = mds_var_dir, 
                                       grid_dir = mds_grid_dir,
                                       read_grid = True,
                                       prefix = mds_files,
                                       geometry = 'llc', 
                                       iters = 'all',
                                       ref_date = ref_date, 
                                       delta_t  = delta_t,
                                       default_dtype = np.dtype(mds_datatype),
                                       grid_vars_to_coords=True)
    
    else:
        if isinstance(model_time_steps_to_load, int):
            iters_to_load = [iters_to_load]
        
        if isinstance(model_time_steps_to_load, list):
            ecco_dataset = open_mdsdataset(data_dir = mds_var_dir, 
                                           grid_dir = mds_grid_dir,
                                           read_grid = True,
                                           prefix = mds_files, 
                                           geometry = 'llc', 
                                           iters = model_time_steps_to_load,
                                           ref_date = ref_date, 
                                           delta_t = delta_t,
                                           default_dtype = np.dtype(mds_datatype),
                                           grid_vars_to_coords=True)
        else:
            print ('not a valid iters_to_load.  must be "all", an "int", or a list of "int"')
            return []


    # replace the xmitgcm coordinate name of 'FACE' with 'TILE'
    if 'face' in ecco_dataset.coords.keys():
        ecco_dataset = ecco_dataset.rename({'face': 'tile'})
        ecco_dataset.tile.attrs['standard_name'] = 'tile_index'
       
    # if vars_to_load is an empty list, keep all variables.  otherwise,
    # only keep those variables in the vars_to_load list.
    
    vars_ignored = []
    vars_loaded = []
    
    if not instance(vars_to_load, list):
        vars_to_load = [vars_to_load]

    if vars_to_load != 'all':
        if isinstance(vars_to_load, str):
            vars_to_load = [vars_to_load]

        print ('loading subset of variables: ', vars_to_load)
    
        # remove variables that are not on the vars_to_load_list
        for ecco_var in ecco_dataset.keys():
            
            if ecco_var not in vars_to_load:
                vars_ignored.append(ecco_var)
                ecco_dataset = ecco_dataset.drop(ecco_var)
            
            else:
                vars_loaded.append(ecco_var)

        print ('loaded  : ', vars_loaded)
        print ('ignored : ', vars_ignored)
    
    else:
        print ('loaded all variables  : ', ecco_dataset.keys())
        
    # keep tiles in the 'tiles_to_load' list.
    if not instance(tiles_to_load, list):
        tiles_to_load = [tiles_to_load]

    ecco_dataset = ecco_dataset.sel(tile = tiles_to_load)
    
    #ecco_dataset = ecco_dataset.isel(time=0)
    if 'AVG' in output_freq_code and \
        'time_bnds' not in ecco_dataset.keys():

        time_bnds_ds, center_times = \
            make_time_bounds_and_center_times_from_ecco_dataset(ecco_dataset,\
                                                                output_freq_code)
        
        if isinstance(ecco_dataset.time.values, np.datetime64):
            ecco_dataset.time.values = center_times
        
        elif isinstance(center_times, np.datetime64):
            center_times = np.array(center_times)
            ecco_dataset.time.values[:] = center_times
        
        if 'ecco-v4-time-average-center-no-units' in meta_common:
            ecco_dataset.time.attrs = \
                meta_common['ecco-v4-time-average-center-no-units']
          
        ecco_dataset = xr.merge((ecco_dataset, time_bnds_ds))
        
        if 'time_bnds-no-units' in meta_common:
            ecco_dataset.time_bnds.attrs=meta_common['time_bnds-no-units']
      
        ecco_dataset = ecco_dataset.set_coords('time_bnds')
        
    elif  'SNAPSHOT' in output_freq_code:
         if 'ecco-v4-time-snapshot-no-units' in meta_common:
            ecco_dataset.time.attrs = \
                meta_common['ecco-v4-time-snapshot-no-units']
       

    #%% DROP SOME EXTRA FIELDS THAT DO NOT NEED TO BE IN THE DATASET
    if 'maskCtrlS' in ecco_dataset.coords.keys():
        ecco_dataset=ecco_dataset.drop('maskCtrlS')
    if 'maskCtrlW' in ecco_dataset.coords.keys():
        ecco_dataset=ecco_dataset.drop('maskCtrlW')
    if 'maskCtrlC' in ecco_dataset.coords.keys():
        ecco_dataset=ecco_dataset.drop('maskCtrlC')
        
    # UPDATE THE VARIABLE SPECIFIC METADATA USING THE 'META_VARSPECIFIC' DICT.
    # if it exists
    for ecco_var in ecco_dataset.variables.keys():
        if ecco_var in meta_variable_specific.keys():
            ecco_dataset[ecco_var].attrs = meta_variable_specific[ecco_var]


    #%% UPDATE THE GLOBAL METADATA USING THE 'META_COMMON' DICT, if it exists
    ecco_dataset.attrs = dict()

    if 'ecco-v4-global' in meta_common:
        ecco_dataset.attrs.update(meta_common['ecco-v4-global'])

    if 'k' in ecco_dataset.dims.keys() and \
        'ecco-v4-global-3D' in meta_common:
        ecco_dataset.attrs.update(meta_common['ecco-v4-global-3D'])
   
    ecco_dataset.attrs['date_created'] = time.ctime()
    
    # give it a hug?
    ecco_dataset = ecco_dataset.squeeze()

    return ecco_dataset




def read_llc_to_tiles(fdir, fname, llc=90, skip=0, nk=1, nl=1, 
                      filetype = '>f4'):
    """
    Loads an MITgcm binary file in the 'tiled' format of the 
    lat-lon-cap (LLC) grids via xmitgcm.  

    Array is returned with the following dimension order:

        [N_tiles, N_recs, N_z, llc, llc]

    where if either N_z or N_recs =1, then that dimension is collapsed
    and not present in the returned array.

    Parameters
    ----------
    fdir : string
        A string with the directory of the binary file to open
    
    fname : string
        A string with the name of the binary file to open
    
    llc : int, optional, default 90
        the size of the llc grid.  For ECCO v4, we use the llc90 domain 
        so `llc` would be `90`.  
        Default: 90
    
    skip : int, optional, default 0
        the number of 2D slices (or records) to skip.  Records could be vertical levels of a 3D field, or different 2D fields, or both.
    
    nk : int, optional, default 1 [singleton]
        number of 2D slices (or records) to load in the depth dimension.  
        Default: 1 [singleton]
    
    nl : int, optional, default 1 [singleton]
        number of 2D slices (or records) to load in the "record" dimension.  
        Default: 1 [singleton] 
    
    filetype: string, default '>f4'
        the file type, default is big endian (>) 32 bit float (f4)
        alternatively, ('<d') would be little endian (<) 64 bit float (d)
        
    Returns
    -------
    data_tiles : ndarray
        a numpy array of dimension 13 x nl x nk x llc x llc, one llc x llc array 
        for each of the 13 tiles and nl and nk levels.  

    """

    full_filename = '%s/%s' % (fdir,fname)

    # Handle "skipped" records by reading up until that record, and
    # dropping preceding records afterward
    #
    # Note: that xmitgcm looks at recs including a full 3D chunk
    # while "skip" refers to 2D chunks.
    nrecs = nl
    skip_3d = int(skip/nk)
    nrecs += skip_3d

    # Reads data into dask array as numpy memmap 
    # [Nrecs x Nz x Ntiles x llc x llc]
    data_tiles = xmitgcm.utils.read_3d_llc_data(full_filename, nx=llc, nz=nk,
                                                nrecs=nrecs, dtype=filetype)

    # Handle cases of single or multiple records, and skip>0
    # Also, swap so that Ntiles dim is ALWAYS first 
    # for ecco_v4_py convention
    if nl==1:
        # Only want 1 record
        data_tiles = data_tiles[skip_3d,...]
        if nk>1:
            data_tiles = data_tiles.swapaxes(0,1)

    else:
        # Want more than one record
        data_tiles = data_tiles[skip_3d:skip_3d+nl,...]

        if nk>1:
            data_tiles = data_tiles.swapaxes(1,2)

        data_tiles = data_tiles.swapaxes(0,1)

    # return the array
    return data_tiles

#%%
def read_llc_to_compact(fdir, fname, llc=90, skip=0, nk=1, nl=1, 
            filetype = '>f4', less_output = False ):
    """

    Loads an MITgcm binary file in the 'tiled' format of the 
    lat-lon-cap (LLC) grids, then converts to the compact form

    Array is returned with the following dimension order:

        [N_recs, N_z, N_tiles*llc, llc]

    where if either N_z or N_recs =1, then that dimension is collapsed 
    and not present in the returned array.

    Parameters
    ----------
    fdir : string
        A string with the directory of the binary file to open
    
    fname : string
        A string with the name of the binary file to open
    
    llc : int, optional, default 90
        the size of the llc grid.  For ECCO v4, we use the llc90 domain 
        so `llc` would be `90`.  
        Default: 90
    
    skip : int, optional, default 0
        the number of 2D slices (or records) to skip.  Records could be vertical levels of a 3D field, or different 2D fields, or both.
    
    nk : int, optional, default 1 [singleton]
        number of 2D slices (or records) to load in the depth dimension.  
        Default: 1 [singleton]
    
    nl : int, optional, default 1 [singleton]
        number of 2D slices (or records) to load in the "record" dimension.  
        Default: 1 [singleton] 
    
    filetype: string, default '>f4'
        the file type, default is big endian (>) 32 bit float (f4)
        alternatively, ('<d') would be little endian (<) 64 bit float (d)
        
    less_output : boolean, optional, default False
        A debugging flag.  False = less debugging output
        
        
    Returns
    -------
    data_compact : ndarray
        a numpy array of dimension nl x nk x 13*llc x llc 

    """

    data_tiles = read_llc_to_tiles(fdir,fname,llc=llc,nk=nk,nl=nl,skip=skip,
                                   filetype=filetype)

    data_compact = llc_tiles_to_compact(data_tiles,less_output=less_output)
    
    # return the array
    return data_compact




def read_llc_to_faces(fdir, fname, llc=90, skip=0, nk=1, nl=1,
        filetype = '>f4', less_output = False):
    """

    Loads an MITgcm binary file in the 'compact' format of the 
    lat-lon-cap (LLC) grids and converts it to the '5 faces' format
    of the LLC grids.  

    Can load 2D and 3D arrays.

    Array is returned with the following dimension order:
    - [N_faces][N_recs, N_z, N_y, N_x]

    where if either N_z or N_recs =1, then that dimension is collapsed
    and not present in the returned array.

    Parameters
    ----------
    fdir : string
        A string with the directory of the binary file to open
    
    fname : string
        A string with the name of the binary file to open
    
    llc : int, optional, default 90
        the size of the llc grid.  For ECCO v4, we use the llc90 domain (llc=90)
    
    skip : int, optional, default 0
        the number of 2D slices (or records) to skip.  Records could be vertical levels of a 3D field, or different 2D fields, or both.
    
    nk : int, optional, default 1 [singleton]
        number of 2D slices (or records) to load in the depth dimension.  
    
    nl : int, optional, default 1 [singleton]
        number of 2D slices (or records) to load in the "record" dimension.  
    
    filetype : string, default '>f4'
        the file type, default is big endian (>) 32 bit float (f4)
        alternatively, ('<d') would be little endian (<) 64 bit float (d)
        
    less_output : boolean, optional, default False
        A debugging flag.  False = less debugging output
        

    Returns
    -------
    data_faces : dict
        a dictionary containing the five lat-lon-cap faces data_faces[n]
        dimensions of each 2D slice of data_faces
        f1,f2: 3*llc x llc
        f3: llc x llc
        4,f5: llc x 3*llc  

    """
    
    data_tiles = read_llc_to_tiles(fdir,fname,llc=llc,nk=nk,nl=nl,skip=skip,
                                   filetype=filetype)

    data_faces = llc_tiles_to_faces(data_tiles, less_output = less_output)

    return data_faces






from .ecco_utils import make_time_bounds_and_center_times_from_ecco_dataset
from .ecco_utils import make_time_bounds_from_ds64
from .ecco_utils import extract_yyyy_mm_dd_hh_mm_ss_from_datetime64
from .ecco_utils import createShapefileFromXY
from .ecco_utils import minimal_metadata
from .ecco_utils import months2days

#from extract_grid_fields_from_mitgrid_files import extract_U_point_grid_fields_from_mitgrid_as_tiles
#from extract_grid_fields_from_mitgrid_files import extract_G_point_grid_fields_from_mitgrid_as_tiles

from .llc_array_conversion  import llc_compact_to_tiles
from .llc_array_conversion  import llc_tiles_to_compact
from .llc_array_conversion  import llc_compact_to_faces
from .llc_array_conversion  import llc_faces_to_tiles
from .llc_array_conversion  import llc_tiles_to_faces
from .llc_array_conversion  import llc_faces_to_compact

from .netcdf_product_generation import create_nc_grid_files_on_native_grid_from_mds
from .netcdf_product_generation import get_time_steps_from_mds_files
from .netcdf_product_generation import create_nc_variable_files_on_native_grid_from_mds
from .netcdf_product_generation import update_ecco_dataset_geospatial_metadata
from .netcdf_product_generation import update_ecco_dataset_temporal_coverage_metadata


from .read_bin_llc import read_llc_to_tiles 
from .read_bin_llc import read_llc_to_compact
from .read_bin_llc import read_llc_to_faces
from .read_bin_llc import load_ecco_vars_from_mds

from .read_bin_gen import load_binary_array

from .resample_to_latlon import resample_to_latlon

from .tile_exchange import append_border_to_tile
from .tile_exchange import add_borders_to_DataArray_V_points
from .tile_exchange import add_borders_to_DataArray_U_points
from .tile_exchange import add_borders_to_DataArray_G_points
from .tile_exchange import get_llc_tile_border_mapping
from .tile_exchange import add_borders_to_GRID_tiles

#from .tile_io import load_all_tiles_from_netcdf
#from .tile_io import load_tile_from_netcdf
#from .tile_io import load_subset_tiles_from_netcdf
from .tile_io import load_ecco_grid_tiles_from_nc
from .tile_io import recursive_load_ecco_var_tiles_from_nc
from .tile_io import load_ecco_var_tiles_from_nc

from .tile_plot import plot_tile
from .tile_plot import plot_tiles
from .tile_plot import unique_color

from .tile_plot_proj import plot_proj_to_latlon_grid
from .tile_plot_proj import plot_pstereo
from .tile_plot_proj import plot_global


from .tile_rotation import reorient_13_tile_GRID_Dataset_to_latlon_layout
from .tile_rotation import reorient_13_tile_Dataset_to_latlon_layout_CG_points
from .tile_rotation import rotate_single_tile_Dataset_CG_points
from .tile_rotation import rotate_single_tile_DataArray_CG_points
from .tile_rotation import reorient_13_tile_Dataset_to_latlon_layout_UV_points
from .tile_rotation import rotate_single_tile_Datasets_UV_points
from .tile_rotation import rotate_single_tile_DataArrays_UV_points

from .test_llc_array_loading_and_conversion import run_read_bin_and_llc_conversion_test


__all__ = ['ecco_utils', 
           'llc_array_conversion', 
           'netcdf_product_generation',
           'read_bin_llc',
           'read_bin_gen', 
           'resample_to_latlon', 
           'tile_exchange', 
           'tile_io', 
           'tile_plot',
           'tile_plot_proj', 
           'tile_rotation', 
           'test_llc_array_loading_and_conversion']



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""ECCO v4 Python: Dataset Utililites

This module includes utility routines to create the ECCO product as netcdf and
for processing metadata.

.. _ecco_v4_py Documentation :
   https://github.com/ECCO-GROUP/ECCOv4-py
"""

from __future__ import division, print_function
import numpy as np
import xarray as xr
import datetime
import time
import xmitgcm as xm
import dateutil
import glob
import os


from .read_bin_llc import load_ecco_vars_from_mds
from .ecco_utils import extract_yyyy_mm_dd_hh_mm_ss_from_datetime64

#%%    
def create_nc_grid_files_on_native_grid_from_mds(grid_input_dir, 
                                                grid_output_dir, 
                                                meta_variable_specific = [],
                                                meta_common = [], 
                                                title_basename='ECCOv4r3_grid_tile_',
                                                title='ECCOv4R3 MITgcm grid information',
                                                mds_datatype = '>f4'):
    
    mds_files = ''
#    iters_to_load = []
#    vars_to_load = []
    
    grid =  load_ecco_vars_from_mds(grid_input_dir, 
                                    mds_files, 
                                    grid_input_dir,
                                    meta_variable_specific = meta_variable_specific,
                                    meta_common = meta_common,
                                    mds_datatype = mds_datatype)

    print(grid)
    
    for key in grid.attrs.keys():
        if 'geo' in key or 'time' in key or 'cell_method' in key:
            grid.attrs.pop(key)
            
    grid.attrs['title'] = title
    
    # create the path if it does not exist/
    if not os.path.exists(grid_output_dir):
        os.makedirs(grid_output_dir)

    for i in range(13):
        tmp = grid.sel(tile=i)
        new_fname = grid_output_dir + '/' + title_basename + \
                      str(i).zfill(2) + '.nc'
        print (new_fname)
        
        tmp.to_netcdf(new_fname)
        
    return grid



#%%
def get_time_steps_from_mds_files(mds_var_dir, mds_file):
   
    mds_search = mds_var_dir + '/' + mds_file + '*meta'
    
    tmp_files = np.sort(glob.glob(mds_search))
    
    time_steps = []
    
    for i in range(len(tmp_files)):
        time_step =  int(tmp_files[i][-15:-5])
        time_steps.append(time_step)
    
    return time_steps
#%%

def create_nc_variable_files_on_native_grid_from_mds(mds_var_dir, 
                                                     mds_files_to_load,
                                                     mds_grid_dir,
                                                     output_dir,
                                                     output_freq_code,
                                                     vars_to_load = [],
                                                     tiles_to_load = range(13),
                                                     time_steps_to_load = [],
                                                     meta_variable_specific = dict(),
                                                     meta_common = dict(),
                                                     mds_datatype = '>f4'):

    # force mds_files_to_load to be a list (if str is passed)
    if isinstance(mds_files_to_load, str):
        mds_files_to_load = [mds_files_to_load]

    # force time_steps_to_load to be a list (if int is passed)
    if isinstance(time_steps_to_load, int):
        time_steps_to_load = [time_steps_to_load]

    # for ce tiles_to_load to be a list (if int is passed)
    if isinstance(tiles_to_load, int):
        tiles_to_load = [tiles_to_load]
        
    # loop through each mds file in mds_files_to_load
    for mds_file in mds_files_to_load:
        
        # if time steps to load is empty, load all time steps
        if len(time_steps_to_load ) == 0:
            # go through each file, pull out the time step, add the time step to a list,
            # and determine the start and end time of each record.
           
           time_steps_to_load = \
               get_time_steps_from_mds_files(mds_var_dir, mds_file)

            
        #print ('mds_file ', mds_file)
        
        # find all of the files in this directory matching the name 'mds_file'
        #mds_search = mds_var_dir + '/' + mds_file + '*meta'
        
        #print ('searching in ', mds_search)
        #tmp_files = np.sort(glob.glob(mds_search))
            
        #if len(tmp_files) == 0:
        #    print ('no files matching ', mds_file)
        #    break
        
        first_meta_fname  = mds_file + '.' + \
            str(time_steps_to_load[0]).zfill(10) + '.meta'    

       
        # get metadata for the first file and determine which variables
        # are present
        meta = xm.utils.parse_meta_file(mds_var_dir + '/' + first_meta_fname)
        vars_here =  meta['fldList']
        
        # verify that at least one of the vars_to_load is present in vars_here
        if len(vars_to_load) > 0:
            num_vars_matching = len(np.intersect1d(vars_to_load, vars_here))
        else:
            num_vars_matching = 1
        
        # only proceed if we are sure that the variable we want is in this
        # mds file
        if num_vars_matching == 0:
            print ('none of the variables you want are in ', mds_file)
            break
        
        # loop through time steps
        for time_step in time_steps_to_load:
            #print (mds_file, time_step)
 
            # load the dataset
            ecco_dataset =  \
                load_ecco_vars_from_mds(mds_var_dir, \
                                        mds_file,
                                        mds_grid_dir,
                                        vars_to_load = vars_to_load,
                                        tiles_to_load=tiles_to_load,
                                        iters_to_load=[time_step],
                                        output_freq_code = \
                                              output_freq_code, 
                                        meta_variable_specific = \
                                              meta_variable_specific,
                                        meta_common=meta_common,
                                        mds_datatype=mds_datatype)
                                        #ignore_unknown_vars)
            
            # pull out the associated time
            year, mon, day, hh, mm, ss  = \
                 extract_yyyy_mm_dd_hh_mm_ss_from_datetime64(ecco_dataset.time.values)
                        
            print(year,mon,day)
            
            # if the field comes from an average, 
            # extract the time bounds -- we'll use it before we save
            # the variable
            
            if 'AVG' in output_freq_code:
                tb = ecco_dataset.time_bnds
                tb.name = 'tb'
                
            # loop through each variable in this dataset, 
            for var in ecco_dataset.keys():
                print ('    ' + var)                
                var_ds = ecco_dataset[var]
                
#                # give the variable name a little 
#                if output_freq_code   == 'AVG_MON':
#                    var_ds.name = var_ds.name + '_mon'
#                elif output_freq_code == 'AVG_DAY':
#                    var_ds.name = var_ds.name + '_day'
#                elif output_freq_code == 'AVG_WEEK':
#                    var_ds.name = var_ds.name + '_wk'
#                elif output_freq_code == 'AVG_YEAR':
#                    var_ds.name = var_ds.name + '_yr'
#                elif output_freq_code == 'SNAPSHOT_MON':
#                    var_ds.name = var_ds.name + '_inst'
#                else:
#                    print ('invalid output frequency code specified')
#                    print ('leaving the variable name unchanged')
                    
                # drop these ancillary fields -- they are in grid anyway    
                for key in var_ds.coords.keys():
                    if 'CS' in key or 'SN' in key or 'Depth' in key:
                        var_ds = var_ds.drop(key)
                    
                # save each tile separately
                for tile_i in range(13):
                
                    # pull out the tile
                    tmp = var_ds.isel(tile=tile_i)
                    
                    # create the new file path name
                    if 'MON' in output_freq_code:

                        fname = var + '_' + \
                                str(year) + '_' + \
                                str(mon).zfill(2) + '_tile_' + \
                                str(tile_i).zfill(2) + '.nc'
            
                        newpath = output_dir + '/' + var + '/' + \
                            str(year) + '/' + str(mon).zfill(2)
                    
                    elif ('WEEK' in output_freq_code) or \
                         ('DAY' in output_freq_code):
                                                
                        fname = var + '_' + \
                                str(year) + '_' + \
                                str(mon).zfill(2) + '_' + \
                                str(day).zfill(2) + '_tile_' + \
                                str(tile_i).zfill(2) + '.nc'
                        d0 = datetime.datetime(year, 1,1)
                        d1 = datetime.datetime(year, mon, day)
                        doy = (d1-d0).days + 1
                        
                        newpath = output_dir + '/' + var + '/' + \
                            str(year) + '/' + str(doy).zfill(3)
                   
                        #print (d0, d1)
                        
                    elif 'YEAR' in output_freq_code:
                      
                         fname = var + '_' + \
                                str(year) + '_' + '_tile_' + \
                                str(tile_i).zfill(2) + '.nc'
        
                         newpath = output_dir + '/' + var + '/' + \
                            str(year) 
                        
                    else:
                        print ('no valid output frequency code specified')
                        print ('saving to year/mon/day/tile')
                        fname = var + '_' + \
                            str(year) + '_' + \
                            str(mon).zfill(2) + '_' + \
                            str(day).zfill(2) + '_tile_' + \
                            str(tile_i).zfill(2) + '.nc'
                        
                        newpath = output_dir + '/' + var + '/' + \
                            str(year) + '/' + str(doy).zfill(3)
                        
    
                    # create the path if it does not exist/
                    if not os.path.exists(newpath):
                        os.makedirs(newpath)
                 
                    # convert the data array to a dataset.
                    tmp = tmp.to_dataset()

                    # add the time bounds field back in if we have an 
                    # average field
                    if 'AVG' in output_freq_code:
                        tmp = xr.merge((tmp, tb))
                        tmp = tmp.drop('tb')
    
                    # put the metadata back in
                    tmp.attrs = ecco_dataset.attrs
                    
                    # update the temporal and geospatial metadata
                    tmp = update_ecco_dataset_geospatial_metadata(tmp)
                    tmp = update_ecco_dataset_temporal_coverage_metadata(tmp)
                    
                    # save to netcdf.  it's that simple.
                    tmp.to_netcdf(newpath + '/' + fname)
                    
    return ecco_dataset, tmp


   
#%%
def update_ecco_dataset_temporal_coverage_metadata(ecco_dataset):
    """

    Adds high-level temporal coverage metadata to dataset object if the 
    dataset object has 'time_bnds' coordinates
    
    Input
    ----------    
    ecco_dataset : an xarray dataset


    Output: 
    ----------    
    ecco_dataset : dataset updated with 'time_coverage_start/end', if such
    bounds can be determined
    
    """
    
    if 'time_bnds' in ecco_dataset.coords.keys():
        
        # if there is only one time bounds
        if len(ecco_dataset.time_bnds.shape) == 1:
            ecco_dataset.attrs['time_coverage_start'] = \
                  str(ecco_dataset.time_bnds.values[0])[0:19]
            ecco_dataset.attrs['time_coverage_end']   = \
                    str(ecco_dataset.time_bnds.values[1])[0:19]

        else:
        # if there are many time bounds 
            ecco_dataset.attrs['time_coverage_start'] = \
                str(ecco_dataset.time_bnds.values[0][0])[0:19]
            ecco_dataset.attrs['time_coverage_end']   = \
                str(ecco_dataset.time_bnds.values[-1][-1])[0:19]
    #elif 'time' in ecco_dataset.coords.keys():
    #    ecco_dataset.attrs['time_coverage_start'] = str(ecco_dataset.time.values[0])[0:19]
    #    ecco_dataset.attrs['time_coverage_end']   = str(ecco_dataset.time.values[-1])[0:19]
    
    return ecco_dataset

#%%
def update_ecco_dataset_geospatial_metadata(ecco_dataset):
    """

    Adds high-level geographical coverage metadata to dataset object if the 
    dataset object has 'YG or YC' coordinates
    
    Input
    ----------    
    ecco_dataset : an xarray dataset


    Output: 
    ----------    
    ecco_dataset : dataset updated with 'geospatial extents', if such
    bounds can be determined
    
    """
    
        # set geospatial bounds
    if 'YG' in ecco_dataset.coords.keys() :
        ecco_dataset.attrs['geospatial_lat_max'] = ecco_dataset.YG.values.max()
        ecco_dataset.attrs['geospatial_lat_min'] = ecco_dataset.YG.values.min()     
        ecco_dataset.attrs['nx'] = ecco_dataset.YG.shape[-2]
        ecco_dataset.attrs['ny'] = ecco_dataset.YG.shape[-1]
    
    elif 'YC' in ecco_dataset.coords.keys() :
        ecco_dataset.attrs['geospatial_lat_max'] = ecco_dataset.YC.values.max()
        ecco_dataset.attrs['geospatial_lat_min'] = ecco_dataset.YC.values.min()     
        ecco_dataset.attrs['nx'] = ecco_dataset.YC.shape[-2]
        ecco_dataset.attrs['ny'] = ecco_dataset.YC.shape[-1]
    
    if 'XG' in ecco_dataset.coords.keys():
        ecco_dataset.attrs['geospatial_lon_max'] = ecco_dataset.XG.values.max()
        ecco_dataset.attrs['geospatial_lon_min'] = ecco_dataset.XG.values.min()
    elif 'XC' in ecco_dataset.coords.keys():
        ecco_dataset.attrs['geospatial_lon_max'] = ecco_dataset.XC.values.max()
        ecco_dataset.attrs['geospatial_lon_min'] = ecco_dataset.XC.values.min()
            
    if 'drF' in ecco_dataset.coords.keys():
        ecco_dataset.attrs['geospatial_vertical_max'] = \
            ecco_dataset.drF.cumsum().values[-1]
        ecco_dataset.attrs['geospatial_vertical_min'] = 0
        ecco_dataset.attrs['nz'] = len(ecco_dataset.k)
    else:
        ecco_dataset.attrs['geospatial_vertical_max'] = 0
        ecco_dataset.attrs['geospatial_vertical_min'] = 0
        ecco_dataset.attrs['nz'] = 1
    
    return ecco_dataset


"""
ECCO v4 Python: tile_plot_proj

This module includes routines for plotting arrays in different
projections.

.. _ecco_v4_py Documentation :
   https://github.com/ECCO-GROUP/ECCOv4-py

"""

from __future__ import division,print_function
import numpy as np
import matplotlib.pylab as plt
import matplotlib.path as mpath
import cartopy.crs as ccrs
import cartopy.feature as cfeature
from .resample_to_latlon import resample_to_latlon


#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
def plot_proj_to_latlon_grid(lons, lats, data, 
                             projection_type = 'robin', 
                             plot_type = 'pcolormesh', 
                             user_lon_0 = -66,
                             lat_lim = 50, 
                             levels = 20, 
                             cmap='jet', 
                             dx=.25, 
                             dy=.25,
                             show_colorbar = False, 
                             show_grid_lines = True,
                             show_grid_labels = True,
                             **kwargs):
    
    #%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    # default projection type = robinson
    # default central longitude = 60W
    # default no colorbar, no grid labels, no grid lines.
    # default color limits take the min and max of the values
    # default plot_type is pcolormesh.
    # default lat/lon spacing in lat/lon grid is 0.25 degrees
    # default number of levels for contourf is 20 (levels)
    # default latitude limit for polar stereographic plots is 50N (lat_lim)
    # default colormap is 'jet'

    #%%    
    cmin = np.nanmin(data)
    cmax = np.nanmax(data)

    for key in kwargs:
        if key == "cmin":
            cmin = kwargs[key]
        elif key == "cmax":
            cmax =  kwargs[key]
        else:
            print("unrecognized argument ", key)



    #%%
    # To avoid plotting problems around the date line, lon=180E, -180W 
    # plot the data field in two parts, A and B.  
    # part 'A' spans from starting longitude to 180E 
    # part 'B' spans the from 180E to 360E + starting longitude.  
    # If the starting  longitudes or 0 or 180 it is a special case.
    if user_lon_0 > -180 and user_lon_0 < 180:
        A_left_limit = user_lon_0
        A_right_limit = 180
        B_left_limit =  -180
        B_right_limit = user_lon_0
        center_lon = A_left_limit + 180
        
    elif user_lon_0 == 180 or user_lon_0 == -180:
        A_left_limit = -180
        A_right_limit = 0
        B_left_limit =  0
        B_right_limit = 180
        center_lon = 0
    else:
        raise ValueError('invalid starting longitude')

    #%%
    # the number of degrees spanned in part A and part B
    num_deg_A =  int((A_right_limit - A_left_limit)/dx)
    num_deg_B =  int((B_right_limit - B_left_limit)/dx)

    # find the longitudal limits of part A and B
    lon_tmp_d = dict()
    if num_deg_A > 0:
        lon_tmp_d['A'] = [A_left_limit, A_right_limit]
            
    if num_deg_B > 0:
        lon_tmp_d['B'] = [B_left_limit, B_right_limit]

    # Make projection axis
    if projection_type == 'Mercator':
        ax = plt.axes(projection =  ccrs.Mercator(central_longitude=user_lon_0))

    elif projection_type == 'PlateCaree':
        ax = plt.axes(projection = ccrs.PlateCarree(central_longitude=user_lon_0))

    elif projection_type == 'cyl':
        ax = plt.axes(projection = ccrs.LambertCylindrical(central_longitude=user_lon_0))
        print ('Cannot label gridlines on a LambertCylindrical plot.  Only PlateCarree and Mercator plots are currently supported.')        
        show_grid_labels = False

    elif projection_type == 'robin':    
        ax = plt.axes(projection = ccrs.Robinson(central_longitude=user_lon_0))
        show_grid_labels=False
        print ('Cannot label gridlines on a Robinson plot.  Only PlateCarree and Mercator plots are currently supported.')
        show_grid_labels = False

    elif projection_type == 'ortho':
        ax = plt.axes(projection =  ccrs.Orthographic(central_longitude=user_lon_0))
        print ('Cannot label gridlines on a Orthographic plot.  Only PlateCarree and Mercator plots are currently supported.')        
        show_grid_labels = False

    elif projection_type == 'stereo':    
        if lat_lim > 0:
            ax = plt.axes(projection =ccrs.NorthPolarStereo())
        else:
            ax = plt.axes(projection =ccrs.SouthPolarStereo())

        print ('Cannot label gridlines on a polar stereographic plot.  Only PlateCarree and Mercator plots are currently supported.')            
        show_grid_labels = False

    elif projection_type == 'InterruptedGoodeHomolosine':
        print ('Cannot label gridlines on a InterruptedGoodeHomolosine plot.  Only PlateCarree and Mercator plots are currently supported.')            
        
        ax = plt.axes(projection = ccrs.InterruptedGoodeHomolosine(central_longitude=user_lon_0))
        show_grid_labels = False
        
    else:
        raise ValueError('projection type must be either "Mercator", "PlateCaree",  "cyl", "robin", "ortho", or "stereo"')

    print ('projection type ', projection_type)
    

    #%%
    # loop through different parts of the map to plot (if they exist), 
    # do interpolation and plot
    f = plt.gcf()
    
    print(len(lon_tmp_d))
    for key, lon_tmp in lon_tmp_d.items():

        new_grid_lon, new_grid_lat, data_latlon_projection = \
            resample_to_latlon(lons, lats, data, 
                               -90+dy, 90-dy, dy,
                               lon_tmp[0], lon_tmp[1], dx, 
                               mapping_method='nearest_neighbor')
            
        if isinstance(ax.projection, ccrs.NorthPolarStereo) or \
           isinstance(ax.projection, ccrs.SouthPolarStereo) :
            p, gl, cbar = \
                plot_pstereo(new_grid_lon,
                             new_grid_lat, 
                             data_latlon_projection,
                             4326, lat_lim, 
                             cmin, cmax, ax,
                             plot_type = plot_type,
                             show_colorbar=False, 
                             circle_boundary=True,
                             cmap=cmap, 
                             show_grid_lines=False)

        else: # not polar stereo
            p, gl, cbar = \
                plot_global(new_grid_lon,
                            new_grid_lat, 
                            data_latlon_projection,
                            4326, 
                            cmin, cmax, ax,
                            plot_type = plot_type,                                       
                            show_colorbar = False,
                            cmap=cmap, 
                            show_grid_lines = False,
                            show_grid_labels = False)
                    
        if show_grid_lines :
            ax.gridlines(crs=ccrs.PlateCarree(), 
                                  linewidth=1, color='black', 
                                  alpha=0.5, linestyle='--', 
                                  draw_labels = show_grid_labels)
        
    if show_colorbar:
        sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(cmin,cmax))
        sm._A = []
        cbar = plt.colorbar(sm,ax=ax)        
    
    #%%
    ax.add_feature(cfeature.LAND)
    ax.add_feature(cfeature.COASTLINE)

    ax= plt.gca()

    #%%
    return f, ax, p, cbar
    #%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    
    

def plot_pstereo(xx,yy, data, 
                 data_projection_code, \
                 lat_lim, 
                 cmin, cmax, ax, 
                 plot_type = 'pcolormesh', 
                 show_colorbar=False, 
                 circle_boundary = False, 
                 cmap='jet', 
                 show_grid_lines=False,
                 levels = 20):

                            
    if isinstance(ax.projection, ccrs.NorthPolarStereo):
        ax.set_extent([-180, 180, lat_lim, 90], ccrs.PlateCarree())
        print('north')
    elif isinstance(ax.projection, ccrs.SouthPolarStereo):
        ax.set_extent([-180, 180, -90, lat_lim], ccrs.PlateCarree())
        print('south')
    else:
        raise ValueError('ax must be either ccrs.NorthPolarStereo or ccrs.SouthPolarStereo')

    print(lat_lim)
    
    if circle_boundary:
        theta = np.linspace(0, 2*np.pi, 100)
        center, radius = [0.5, 0.5], 0.5
        verts = np.vstack([np.sin(theta), np.cos(theta)]).T
        circle = mpath.Path(verts * radius + center)
        ax.set_boundary(circle, transform=ax.transAxes)

    if show_grid_lines :
        gl = ax.gridlines(crs=ccrs.PlateCarree(), 
                          linewidth=1, color='black', 
                          alpha=0.5, linestyle='--')
    else:
        gl = []

    if data_projection_code == 4326: # lat lon does nneed to be projected
        data_crs =  ccrs.PlateCarree()
    else:
        # reproject the data if necessary
        data_crs=ccrs.epsg(data_projection_code)
    

    p=[]    
    if plot_type == 'pcolormesh':
        p = ax.pcolormesh(xx, yy, data, transform=data_crs, \
                          vmin=cmin, vmax=cmax, cmap=cmap)

    elif plot_type =='contourf':
        p = ax.contourf(xx, yy, data, levels, transform=data_crs,  \
                 vmin=cmin, vmax=cmax, cmap=cmap)

    else:
        raise ValueError('plot_type  must be either "pcolormesh" or "contourf"')

         
    ax.add_feature(cfeature.LAND)
    ax.coastlines('110m', linewidth=0.8)

    cbar = []
    if show_colorbar:
        sm = plt.cm.ScalarMappable(cmap=cmap,norm=plt.Normalize(cmin,cmax))
        sm._A = []
        cbar = plt.colorbar(sm,ax=ax)
    
    return p, gl, cbar

#%%    

def plot_global(xx,yy, data, 
                data_projection_code,
                cmin, cmax, ax, 
                plot_type = 'pcolormesh', 
                show_colorbar=False, 
                cmap='jet', 
                show_grid_lines = True,
                show_grid_labels = True,
                levels=20):

    if show_grid_lines :
        gl = ax.gridlines(crs=ccrs.PlateCarree(), 
                          linewidth=1, color='black', 
                          draw_labels = show_grid_labels,
                          alpha=0.5, linestyle='--')
    else:
        gl = []
        
    if data_projection_code == 4326: # lat lon does nneed to be projected
        data_crs =  ccrs.PlateCarree()
    else:
        data_crs =ccrs.epsg(data_projection_code)
        
    if plot_type == 'pcolormesh':
        p = ax.pcolormesh(xx, yy, data, transform=data_crs, 
                          vmin=cmin, vmax=cmax, cmap=cmap)
    elif plot_type =='contourf':
        p = ax.contourf(xx, yy, data, levels, transform=data_crs,
                        vmin=cmin, vmax=cmax, cmap=cmap)
    else:
        raise ValueError('plot_type  must be either "pcolormesh" or "contourf"') 
                         
    ax.coastlines('110m', linewidth=0.8)
    ax.add_feature(cfeature.LAND)

    cbar = []
    if show_colorbar:
        sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(cmin,cmax))
        sm._A = []
        cbar = plt.colorbar(sm,ax=ax)
    
    return p, gl, cbar



"""
ECCO v4 Python: tile_io

This module provides routines for loading ECCO netcdf files.

.. _ecco_v4_py Documentation :
   https://github.com/ECCO-GROUP/ECCOv4-py

"""
from __future__ import division,print_function
import numpy as np
import xarray as xr
from copy import deepcopy
import glob
import os

#%%
def load_ecco_grid_tiles_from_nc(grid_dir, 
                                 grid_base_name = 'ECCOv4r3_grid_tile_',
                                 tiles_to_load=range(13), 
                                 k_subset = [], 
                                 use_dask = True, 
                                 coords_as_vars = False):
        
    num_loaded_tiles = 0
    g = []
    
    for i in tiles_to_load:
        if use_dask:
            g_i = xr.open_dataset(grid_dir + '/' + grid_base_name + str(i).zfill(2) + '.nc', chunks=90)
        else:
            g_i = xr.open_dataset(grid_dir + '/' + grid_base_name + str(i).zfill(2) + '.nc')
       
        if len(k_subset) > 0:
                g_i = g_i.isel(k=k_subset)
                
        num_loaded_tiles += 1
        
        if num_loaded_tiles == 1:
            g = g_i
        else:
            g = xr.concat((g, g_i ),'tile')
            
#    for key, value in ds.variables.items():
#        if key not in ds.coords and len(ds[key].shape) >= 2:
#            ds.variables[key].attrs['grid_layout'] = 'original llc'
            
        if coords_as_vars:
            g.reset_coords()
            
            
    return g  


#%%

def recursive_load_ecco_var_tiles_from_nc(data_root_dir, 
                                          var_names,
                                          tiles_to_load = range(13),
                                          k_subset = [],
                                          use_dask = True,
                                          less_output = True):
    
    ecco_dataset = []
    ecco_dataset_var = []


    # if 'ALL' is passed as the var_names then search in data_door_dir
    # for variables.  
    if isinstance(var_names, str) and 'ALL' in var_names:
        var_head_dirs = np.sort(glob.glob(data_root_dir + '/*'))
        var_names = []
       
        for var in var_head_dirs:
           var_names.append(str.split(var,'/')[-1])
       
        print ('searching for variables in ', data_root_dir)
        if len(var_names) == 0:
           print ('none found')
        else:
           print ('found ',  var_names)
    
    
    # if only one var_name is passed (say as a string)) then
    # we convert the string to a list
    elif not isinstance(var_names, list):
        var_names = [var_names]
      
        
    for var_name in var_names:
        #if not less_output:
        print (var_name)
            
        ecco_dataset_var = []
     
        sub_dirs = np.sort(([[x[0] for x in os.walk(data_root_dir)]]))
       
        if not less_output :
            print (sub_dirs[0])
                    
        for sub_dir in sub_dirs[0]:
            if not less_output :
                print (sub_dir)
            
            files = np.sort(glob.glob(sub_dir + '/' + var_name + '*nc'))
           
            if len(files) > 0:
                if not less_output:
                    print ('found files here .. loading ', files)
               
                ecco_dataset_tmp = load_ecco_var_tiles_from_nc(sub_dir, 
                                                               var_name, 
                                                               tiles_to_load,
                                                               k_subset, 
                                                               use_dask, 
                                                               less_output)
               
                if len(ecco_dataset_var) == 0:
                    # create the dataset if it hasn't yet been initialized
                    ecco_dataset_var = ecco_dataset_tmp
                else:
                    # concat the new dataset along the time axis
                    ecco_dataset_var = \
                        xr.concat((ecco_dataset_var, ecco_dataset_tmp), 'time')
        
        #if len(ecco_dataset) == 0 and len(ecco_dataset_var) > 0:
        #    ecco_dataset = ecco_dataset_var
        #else:
        #    ecco_dataset = xr.merge((ecco_dataset, ecco_dataset_var))
        if len(ecco_dataset) == 0 and len(ecco_dataset_var) > 0:
            ecco_dataset = ecco_dataset_var
        elif len(ecco_dataset) > 0 and len(ecco_dataset_var) > 0:
            ecco_dataset = xr.merge((ecco_dataset, ecco_dataset_var))
    
    return ecco_dataset


#%%
def load_ecco_var_tiles_from_nc(data_dir, 
                                var_name, 
                                tiles_to_load = range(13), 
                                k_subset = [], 
                                use_dask = True,
                                less_output = True):
    
    num_loaded_tiles = 0
    g = []    
    files = np.sort(glob.glob(data_dir + '/' + var_name + '*nc'))
    
    for f in files:
        
        tile = int(f[-5:-3])
        if tile in tiles_to_load:
            if not less_output:
                print ('loading tile %d' % tile)
                
            if use_dask:
                g_i = xr.open_dataset(f, chunks=90)
            else:
                g_i = xr.open_dataset(f)
           
            if len(k_subset) > 0:
                g_i = g_i.isel(k=k_subset)
                
            num_loaded_tiles += 1

            #print (g_i.tile.values)
            if num_loaded_tiles == 1:
                g = g_i
            else:
                g = xr.concat((g, g_i ),'tile')
                
        elif not less_output:
            print ('not loading %s tile %d ' % (var_name, tile))
          
#    for key, value in ds.variables.items():
#        if key not in ds.coords and len(ds[key].shape) >= 2:
#            ds.variables[key].attrs['grid_layout'] = 'original llc'

    return g  
 



    
